#include <stdlib.h>
#include <math.h>
#include <stdio.h>	

#include "graph.h"



/*
 *  Allocates and returns a new graph.
 *
 *  int n		number of nodes
 *  double value	default value for the edges on the path (1, 2, ..., n)
 */
graph *graph_new(int n, double weight) {
	graph *G = malloc(sizeof(graph));
	G->n = n;
	G->edge = malloc(n * sizeof(double *));
	G->cat = malloc(n * sizeof(category *));
	for (int u = 0; u < n; u++) {
		G->edge[u] = malloc(n * sizeof(double));
		G->cat[u] = malloc(n * sizeof(category));
		for (int v = 0; v < n; v++) {
			if (u < n - 1 && u + 1 == v) {
				G->edge[u][v] = weight;
				G->cat[u][v] = RND;
			} else {
				G->edge[u][v] = INFINITY;
				G->cat[u][v] = NON;
			}
		}
	}
	return G;
}



/*
 *  Frees a graph's memory.
 */
void graph_free(graph *G) {
	if (G) {
		if (G->edge) {
			for (int u = 0; u < G->n; u++) {
				free(G->edge[u]);
			}
			free(G->edge);
		}
		if (G->cat) {
			for (int u = 0; u < G->n; u++) {
				free(G->cat[u]);
			}
			free(G->cat);
		}
		free(G);
	}
}



/*
 *  Updates a graph edge, if the new weight is less than the previous weight.
 */
void graph_update_edge(graph *G, int u, int v, double weight, category cat) {
	if (G->edge[u][v] > weight) {
		G->edge[u][v] = weight;
		G->cat[u][v] = cat;
	}
}



/*
 *  Computes the shortes path from node 1 to node n.
 *
 *  int *path	string array of size n containing the path
 *		All empty fields are set to -1.
 */
double graph_compute_path(graph *G, int *path) {
	// dist[u]: distance from node u to node n
	double *dist = malloc(G->n * sizeof(double));
	// dist[u]: predecessor of node n
	int *pred = malloc(G->n * sizeof(int));
	for (int i = 0; i < G->n; i++) {
		dist[i] = INFINITY;
		pred[i] = -1;
	}
	dist[0] = 0;
	for (int u = 0; u < G->n; u++) {
		for (int v = 0; v < G->n; v++) {
			if (G->edge[u][v] != INFINITY) {
				if (dist[v] > dist[u] + G->edge[u][v]) {
					dist[v] = dist[u] + G->edge[u][v];
					pred[v] = u;
				}
			}
		}
	}
	int j = 0;
	for (int i = G->n - 1; i >= 0; i = pred[i], j++) {
		path[j] = i;
	}
	for ( ; j < G->n; j++) {
		path[j] = -1;
	}
	double entropy = dist[G->n - 1];
	free(dist);
	free(pred);
	return entropy;
}



/*
 *  Prints out a graph.
 */
void graph_print(graph *G) {
	printf("\n=====================================\n");
	for (int u = 0; u < G->n; u++) {
		for (int v = 0; v < G->n; v++) {
			char *cat;
			switch (G->cat[u][v]) {
				case WRD:
					cat = "wrd";
					break;
				case SEQ:
					cat = "seq";
					break;
				case RND:
					cat = "rnd";
					break;
				case KBP:
					cat = "kbp";
					break;
				default:
					cat = "   ";
					break;
			}
			printf("%.1f %s\t", G->edge[u][v], cat);
		}
		printf("\n");
	}	
	printf("=====================================\n\n");
}



/*
 *  Prints out a path.
 *
 *  int n	size of the path array
 */
void path_print(int *path, int n) {
	for (int i = n - 1; i >= 0; i--) {
		if (path[i] >= 0) {
			printf("%d", path[i]);
			if (i > 0) {
				printf("->");
			}
		}
	}
	printf("\n");
}

