#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>

#include "word.h"
#include "keyboard.h"
#include "sequence.h"
#include "graph.h"



// flags for character sets, SC are non-alphanumeric characters
#define FLAG_az 1
#define FLAG_AZ 2
#define FLAG_09 4
#define FLAG_SC 8

// number of non-alphanumeric characters
#define NUM_SC 32



/*
 *  Returns true, iff a character is a non-blank ASCII character.
 */
int printable_ascii(char *str) {
	for (char *c = str; *c != '\0' && *c != '\n'; c++) {
		if (!('!' <= *c && *c <= '~')) {
			return 0;
		}
	}
	return 1;
}



/*
 *  Computes the charset of a string.
 */
int compute_charset(char *str) {
	char flags = 0;
	for (char *c = str; *c != '\0'; c++) {
		if ('a' <= *c && *c <= 'z') {
			flags |= FLAG_az;
		} else if ('A' <= *c && *c <='Z') {
			flags |= FLAG_AZ;
		} else if ('0' <= *c && *c <= '9') {
			flags |= FLAG_09;
		} else {
			flags |= FLAG_SC;
		}
	}

	int charset = 0;
	if (flags & FLAG_az) {
		charset += 26;
	}
	if (flags & FLAG_AZ) {
		charset += 26;
	}
	if (flags & FLAG_09) {
		charset += 10;
	}
	if (flags & FLAG_SC) {
		charset += NUM_SC;
	}

	return charset;
}



/*
 *  Prints the result line for a substring starting at u, ending at v
 */
void print_line(char *str, int u, int v, graph *G) {
	char *cat, *formula;
	switch (G->cat[u][v]) {
		case WRD:
			cat = "wrd";
			formula = "log2(dict)";
			break;
		case SEQ:
			cat = "seq";
			formula = "log2(36) + (len - 1) * log2(3)";
			break;
		case RND:
			cat = "rnd";
			formula = "log2(chset)";
			break;
		case KBP:
			cat = "kbp";
			formula = "log2(47) + (len - 1) * log2(9)";
			break;
		default:
			cat = "non";
			formula = "ERROR";
			break;
	}

	int len = v - u;
	char *buf = malloc(len * sizeof(char));
	strncpy(buf, str + u, len);
	buf[len] = '\0';
	printf("%s\t%s\t%d\t%.2f\t%s\n", buf, cat, len, G->edge[u][v], formula);
	free(buf);
}



/*
 *  Computes and returns the entropy of the whole password.
 */
double compute_entropy(char *word, dictionary *dict, long dict_words) {
	int n = strlen(word);
	int charset = compute_charset(word);
	graph *G = graph_new(n + 1, log(charset) / log(2));
	int *wrdlen = malloc(n * sizeof(int));

	for (char *c = word; *c != '\0'; c++) {
		int seqlen = find_seq(c);
		int kbplen = find_kbp(c);
		int wrdcnt = find_wrd(dict, c, wrdlen);

		for (int i = 3; i <= seqlen; i++) {
			graph_update_edge(G, c - word, c - word + i, rate_seq(*c, i), SEQ);
		}
		for (int i = 3; i <= kbplen; i++) {
			graph_update_edge(G, c - word, c - word + i, rate_kbp(i), KBP);
		}
		for (int i = 0; i < wrdcnt; i++) {
			graph_update_edge(G, c - word, c - word + wrdlen[i], rate_wrd(dict_words), WRD);
		}
	}
	
	int *path = malloc(G->n * sizeof(int));
	double entropy = graph_compute_path(G, path);

	printf("\nsubstr\ttype\tlength\tentropy\t\tformula\n");
	for (int *i = path + n; i > path; i--) {
		if (*i >= 0) {
			print_line(word, *i, *(i - 1), G);
		}
	}

	free(wrdlen);
	free(path);
	graph_free(G);
	return entropy;
}



/*
 *  Generates random words up to length 64 and uses them for tests.
 */
void fuzz(int count, dictionary *t, long dict_words) {
	FILE *urandom = fopen("/dev/urandom", "r");
	int len = fgetc(urandom) & 0x3f;
	char *word = malloc(len + 1);

	for (int j = 0; j < count; j++) {
		for (int i = 0; i < len; i++) {
			int byte;
			if ((byte = fgetc(urandom)) < 0) {
				fprintf(stderr, "ERROR: Reading \"/dev/urandom\" failed.");
				exit(-1);
			}
			word[i] = '+' + (byte & 0x3f);
		}
		word[len] = '\0';
		printf("\n\n\nTesting: %s\n", word);
		compute_entropy(word, t, dict_words);
	}
	free(word);
	fclose(urandom);
}



/*
 *  Prints out the estimated cracking time, assuming one billion attempts per
 *  second.
 */
void estimated_time(double entropy) {
	double seconds = pow(2, entropy) / 1000 / 1000 / 1000;
	if (seconds < 0.000001) {
		printf("Estimated minimum cracking time: %.0f nanoseconds.\n\n",
				seconds * 1000 * 1000 * 1000);
	} else 	if (seconds < 0.001) {
		printf("Estimated minimum cracking time: %.0f microseconds.\n\n",
				seconds * 1000 * 1000);
	} else if (seconds < 1.0) {
		printf("Estimated minimum cracking time: %.0f milliseconds.\n\n",
				seconds * 1000);
	} else if (seconds < 60) {
		printf("Estimated minimum cracking time: %.0f seconds.\n\n",
				seconds);
	} else if (seconds < 3600) {
		printf("Estimated minimum cracking time: %.0f minutes.\n\n",
				round(seconds / 60));
	} else if (seconds < 3600 * 24) {
		printf("Estimated minimum cracking time: %.0f hours.\n\n",
				round(seconds / 3600));
	} else if (seconds < 3600 * 24 * 365.25) {
		printf("Estimated minimum cracking time: %.0f days.\n\n",
				round(seconds / 3600 / 24));
	} else if (seconds < 3600 * 24 * 365.25 * 1000 * 1000 * 1000) {
		printf("Estimated minimum cracking time: %.0f years.\n\n",
				round(seconds / 3600 / 24 / 365.25));
	} else {
		printf("Estimated minimum cracking time: %.3e years.\n\n",
				round(seconds / 3600 / 24 / 365.25));
	}
}



int main(int argc, char **argv) {
	printf("pwcheck v2.0\n\n");

	long dict_words = 0;
	long dict_nodes = 0;
	dictionary *dict = dictionary_new(&dict_words, &dict_nodes);
	double dict_size = sizeof(dictionary) * dict_nodes / 1024.0 / 1024.0;
	printf("Dictionary has %ld nodes for %ld words. Total size: %.1f MiB\n\n",
			dict_nodes, dict_words, dict_size);

	if (argc > 1) {
		fuzz(atoi(argv[1]), dict, dict_words);
		//compute_entropy(argv[1], dict, dict_words);
		dictionary_free(dict);
		return 0;
	}

	while (1) {
		char buf[256];
		printf("Please enter a password: ");
		if (!fgets(buf, 256, stdin)) {
			fprintf(stderr, "ERROR: fgets() failed.\n");
		}
		buf[strlen(buf) - 1] = '\0';

		if (!printable_ascii(buf)) {
			fprintf(stderr, "ERROR: Only ASCII characters accepted.\n\n");
			continue;
		}
		
		double entropy = compute_entropy(buf, dict, dict_words);
		double max_entropy = strlen(buf) * log(94) / log(2);
		printf("\nCombined entropy for \"%s\": %.2f\n", buf, entropy);
		printf("Maximum entropy for this length: %.2f\n", max_entropy);
		estimated_time(entropy);
	}	

	dictionary_free(dict);

	return 0;
}

